@@ -7,7 +7,7 @@
 #
 #   This program is free software: you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
-#   the Free Software Foundation, either version 3 of the License, or
+#   the Free Software Foundation, either version 2 of the License, or
 #   (at your option) any later version.
 #
 #   This program is distributed in the hope that it will be useful,
@@ -61,13 +61,12 @@ function(x, value) {
   if(value %in% c('POSIXlt','POSIXct'))
     value <- c(value,'POSIXt')
 
-  if(value %in% 'Date') {
-    
-  }
-
   attr(x, '.indexCLASS') <- value
   # all index related meta-data will be stored in the index
   # as attributes
+  if(any(value %in% .classesWithoutTZ)) {
+    attr(attr(x,'index'), 'tzone') <- 'UTC'
+  }
   attr(attr(x,'index'), 'tclass') <- value
   x
 }
